package blueeyes.io.iteratee

trait Enumeratee[EO, EI, A] extends (Iteratee[EI, A] => Iteratee[EO, Iteratee[EI, A]])

import Input._
import Iteratee._

trait CoreEnumeratees {
  /**
   * 
   */
  def mapInput[EO, EI, A](m: EO => EI, mi: EI => EO): Enumeratee[EO, EI, A] = new Enumeratee[EO, EI, A] {
    def apply(inner: Iteratee[EI, A]): Iteratee[EO, Iteratee[EI, A]] = {
      inner match {
        case Done(a, r) => Iteratee.returnValue[EO, Iteratee[EI, A]](inner, r.map(mi))
        
        case Next(n) => Iteratee.continueWith { (i: Input[EO]) =>
          apply(n(i.map(m)))
        }
        
        case Error(e, n) => Iteratee.throwErr(e, (i: Input[EO]) => apply(n(i.map(m))))
      }
    }
  }
  
  def transformInput[EO, EI, A](transformer: Iteratee[EO, EI]): Enumeratee[EO, EI, A] = new Enumeratee[EO, EI, A] {
    def apply(inner: Iteratee[EI, A]) = {
      inner match {
        case Done(a, r)  => Iteratee.returnValue[EO, Iteratee[EI, A]](inner)
        
        case Next(n)     => transformer match {
          case Done(ei, r) => transformInput(Iteratee.remainder(r) >> transformer).apply(n(Chunk(ei)))
          case Next(n2)    => Iteratee.continueWith { (eo: Input[EO]) =>
            transformInput(n2(eo))(inner)
          }
          case Error(e, n2) => Iteratee.throwErr(e, (eo: Input[EO]) => transformInput(n2(eo))(inner))
        }
        
        case Error(e, n) => Iteratee.throwErr(e, 
          (eo: Input[EO]) => transformer match {
            case Done(ei, r)  => transformInput(Iteratee.remainder(r) >> transformer).apply(n(Chunk(ei)))
            case Next(n2)     => transformInput(n2(eo))(inner)
            case Error(e, n2) => Iteratee.throwErr(e, (eo: Input[EO]) => transformInput(n2(eo))(inner))
          }
        )
      }
    }
  }
}
object CoreEnumeratees extends CoreEnumeratees

trait ChunkEnumeratees extends ChunkIteratees {
  /** Takes n chunks from the input and feeds them to an inner iterator. 
   * Barring EOF, n chunks are always consumed, even if the inner iterator 
   * does not use all of them or produces an error. Any remainder from the 
   * inner iterator is always discarded.
   * {{{
   * take(5)(iter)
   * }}}
   */
  def take[E, A](n: Int): Enumeratee[E, E, A] = new Enumeratee[E, E, A] {
    private def take0(count: Int, inner: Iteratee[E, A]): Iteratee[E, Iteratee[E, A]] = {
      def takeNext(count: Int, inner: Iteratee[E, A]): (Input[E] => Iteratee[E, Iteratee[E, A]]) = {
        case Empty => take0(count, inner)
        
        case input @ Chunk(c) => inner match {
          case Done(a, r)   => drop(count - 1) >> Iteratee.returnValue[E, Iteratee[E, A]](inner, Empty)
          case Next(n)      => take0(count - 1, n(input))
          case Error(e, n)  => drop(count - 1) >> Iteratee.returnValue[E, Iteratee[E, A]](inner, Empty)
        }
            
        case Eof => Iteratee.throwErr[E, Iteratee[E, A]](InadequateInput("Expected " + count + " more units of input"), takeNext(count, inner))
      }
      
      if (count <= 0) Iteratee.returnValue(inner, Empty)
      else Iteratee.continueWith(takeNext(count, inner))
    }
    
    def apply(inner: Iteratee[E, A]): Iteratee[E, Iteratee[E, A]] = take0(n, inner)
  }
}
object ChunkEnumeratees extends ChunkEnumeratees

trait VectorLikeEnumeratees extends VectorLikeIteratees {
  def take[E[_], T, A](n: Int)(implicit vl: VectorLike[E]): Enumeratee[E[T], E[T], A] = new Enumeratee[E[T], E[T], A] {
    private def take0(count: Int, inner: Iteratee[E[T], A]): Iteratee[E[T], Iteratee[E[T], A]] = {
      def takeNext(count: Int, inner: Iteratee[E[T], A]): (Input[E[T]] => Iteratee[E[T], Iteratee[E[T], A]]) = {
        case Empty => take0(count, inner)
        
        case input @ Chunk(c) => {
          val taken = vl.take(c, count)
          val rest  = Chunk(vl.drop(c, count))
          
          val numTaken = vl.len(taken)
          
          val remainingCount = count - numTaken
          
          inner match {
            case Done(a, r)   => remainder(rest) >> drop(remainingCount) >> Iteratee.returnValue[E[T], Iteratee[E[T], A]](inner, Empty)
            case Next(n)      => remainder(rest) >> take0(remainingCount, n(input))
            case Error(e, n)  => remainder(rest) >> drop(remainingCount) >> Iteratee.returnValue[E[T], Iteratee[E[T], A]](inner, Empty)
          }
        }
        
        case Eof => Iteratee.throwErr[E[T], Iteratee[E[T], A]](InadequateInput("Expected " + count + " more units of input"), takeNext(count, inner))
      }
      
      if (count <= 0) Iteratee.returnValue(inner, Empty)
      else Iteratee.continueWith(takeNext(count, inner))
    }
    
    def apply(inner: Iteratee[E[T], A]): Iteratee[E[T], Iteratee[E[T], A]] = take0(n, inner)
  }
}
object VectorLikeEnumeratees extends VectorLikeEnumeratees

object Enumeratee extends CoreEnumeratees {
  
}